# -*- coding: utf-8 -*-
# Copyright (C) 2020 fjord-technologies
# SPDX-License-Identifier: GPL-3.0-or-later
"""xtrargon_ng.modules.quorum"""

import logging
import requests

from dwho.classes.modules import DWhoModuleBase, MODULES
from httpdis.ext.httpdis_json import HttpReqErrJson
from sonicprobe.libs import network, urisup, xys
from sonicprobe.libs.moresynchro import RWLock

from ..classes.abstract.db import XtrargonDBBase


LOG                        = logging.getLogger('xtrargon_ng.modules.quorum')

DB_SERVERS_SUFFIX          = {'a': 'b',
                              'b': 'a'}

DEFAULT_CNF_FILE           = 'quorum.cnf'
DEFAULT_BACKUPS_STATUS_URI = 'http://:0/backups/status'
DEFAULT_REQUEST_TIMEOUT    = 15

xys.add_callback('quorum.host', network.valid_domain)

class QuorumDB(XtrargonDBBase):
    pass

class QuorumModule(DWhoModuleBase):
    MODULE_NAME = 'quorum'

    LOCK = RWLock()

    # pylint: disable-msg=attribute-defined-outside-init
    def safe_init(self, options):
        self.lock_timeout = self.config['general']['lock_timeout']
        self.request_timeout = DEFAULT_REQUEST_TIMEOUT

        backups_status_uri = DEFAULT_BACKUPS_STATUS_URI

        if self.config.get('quorum'):
            if self.config['quorum'].get('backups_status_uri'):
                backups_status_uri = self.config['quorum']['backups_status_uri']

            if 'request_timeout' in self.config['quorum']:
                self.request_timeout = float(self.config['quorum']['request_timeout'])

        self.backups_status_uri = list(urisup.uri_help_split(backups_status_uri))

    def _build_backups_status_uri(self, host = None):
        uri = list(self.backups_status_uri)
        if self.backups_status_uri[1]:
            uri[1] = list(self.backups_status_uri[1])
        else:
            uri[1] = [None, None, None, None]

        if host and not uri[1][2]:
            uri[1][2] = host

        if uri[1][3] == '0':
            uri[1][3] = self.config['general'].get('listen_port')

        return urisup.uri_help_unsplit(uri)


    BACKUPS_STATUS_QSCHEMA = xys.load("""
    host: !~~callback(quorum.host)
    """)

    def backups_status(self, request):
        params = request.query_params()

        if not isinstance(params, dict):
            raise HttpReqErrJson(400, "invalid arguments type")

        if not xys.validate(params, self.BACKUPS_STATUS_QSCHEMA):
            raise HttpReqErrJson(415, "invalid arguments for command")

        params['host'] = params['host'].lower()
        suffix = params['host'][-1]

        if suffix not in DB_SERVERS_SUFFIX:
            raise HttpReqErrJson(415, "invalid host argument: %r" % params['host'])

        host = "%s%s" % (params['host'][:-1], DB_SERVERS_SUFFIX[suffix])

        if not self.LOCK.acquire_read(self.lock_timeout):
            raise HttpReqErrJson(503, "unable to take LOCK for reading after %s seconds" % self.lock_timeout)

        req = None

        try:
            LOG.info("get backups status on %r", host)

            req = requests.get(self._build_backups_status_uri(host),
                               timeout = self.request_timeout)
            if not req.text:
                return None

            try:
                return req.json()
            except ValueError:
                req.raise_for_status()
        except HttpReqErrJson:
            raise
        except Exception as e:
            LOG.exception(e)
            raise HttpReqErrJson(503, repr(e))
        finally:
            if req:
                req.close()
            self.LOCK.release()


    REPLICA_STATUS_QSCHEMA = xys.load("""
    host: !~~callback(quorum.host)
    """)

    def replica_status(self, request):
        params = request.query_params()

        if not isinstance(params, dict):
            raise HttpReqErrJson(400, "invalid arguments type")

        if not xys.validate(params, self.REPLICA_STATUS_QSCHEMA):
            raise HttpReqErrJson(415, "invalid arguments for command")

        params['host'] = params['host'].lower()
        suffix = params['host'][-1]

        if suffix not in DB_SERVERS_SUFFIX:
            raise HttpReqErrJson(415, "invalid host argument: %r" % params['host'])

        host = "%s%s" % (params['host'][:-1], DB_SERVERS_SUFFIX[suffix])

        if not self.LOCK.acquire_read(self.lock_timeout):
            raise HttpReqErrJson(503, "unable to take LOCK for reading after %s seconds" % self.lock_timeout)

        qdb = None

        try:
            LOG.info("get replica status on %r", host)
            qdb = QuorumDB()
            qdb.init_db(module = self, default_cnf_file = DEFAULT_CNF_FILE, host = host)
            qdb.create_connection()

            return qdb.get_slave_status()
        except HttpReqErrJson:
            raise
        except Exception as e:
            LOG.exception(e)
            raise HttpReqErrJson(503, repr(e))
        finally:
            if qdb:
                qdb.close_connection()
            self.LOCK.release()


if __name__ != "__main__":
    def _start():
        MODULES.register(QuorumModule())
    _start()
